/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 18/10/22.
//

#include "models/viewer.hpp"

#include "fggl/assets/module.hpp"
#include "fggl/data/assimp/module.hpp"

#include "fggl/mesh/mesh.hpp"
#include "fggl/mesh/components.hpp"

#include "fggl/input/camera_input.hpp"

#include "fggl/gfx/phong.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/gfx/paint.hpp"

namespace demo {

	static fggl::entity::EntityID build_model(fggl::entity::EntityManager& manager, fggl::assets::AssetManager *assets, fggl::assets::AssetID assetRef){
		auto model = manager.create();

		manager.add<fggl::math::Transform>(model);
		auto& mesh = manager.add<fggl::mesh::StaticMultiMesh3D>(model);

		auto* meshData = assets->get<fggl::mesh::MultiMesh3D>( assetRef );
		if ( meshData == nullptr ) {
			fggl::debug::warning("loading model did not work!");
		} else {
			mesh.mesh = *meshData;
			mesh.pipeline = "redbook/debug";
		}

		manager.add<fggl::gfx::PhongMaterial>(model);

		return model;
	}

	static void process_camera(fggl::entity::EntityManager& ecs, const fggl::input::Input& input) {
		auto cameras = ecs.find<fggl::gfx::Camera>();
		auto cam = cameras[0];

		auto camTransform = ecs.get<fggl::math::Transform>(cam);
		auto camComp = ecs.get<fggl::gfx::Camera>(cam);

		const glm::vec3 dir = ( camTransform.origin() - camComp.target );
		const glm::vec3 forward = glm::normalize( dir );

		// scroll wheel
		glm::vec3 motion(0.0f);
		float delta = input.mouse.axis( fggl::input::MouseAxis::SCROLL_Y );
		if ( (glm::length( dir ) < 25.0f && delta < 0.0f) || (glm::length( dir ) > 2.5f && delta > 0.0f) )
			motion -= (forward * delta);
		camTransform.origin( camTransform.origin() + motion );

		fggl::input::process_arcball(ecs, input, cam);
	}

	static void setup_camera(fggl::entity::EntityManager& world) {
		auto prototype = world.create();

		// setup camera position/transform
		auto& transform = world.add<fggl::math::Transform>(prototype);
		transform.origin(glm::vec3(10.0f, 3.0f, 10.0f));

		// setup camera components
		world.add<fggl::gfx::Camera>(prototype);

		// interactive camera
		auto& cameraKeys = world.add<fggl::input::FreeCamKeys>(prototype);
		cameraKeys.forward = glfwGetKeyScancode(GLFW_KEY_W);
		cameraKeys.backward = glfwGetKeyScancode(GLFW_KEY_S);
		cameraKeys.left = glfwGetKeyScancode(GLFW_KEY_A);
		cameraKeys.right = glfwGetKeyScancode(GLFW_KEY_D);
		cameraKeys.rotate_cw = glfwGetKeyScancode(GLFW_KEY_Q);
		cameraKeys.rotate_ccw = glfwGetKeyScancode(GLFW_KEY_E);
	}

	static void setup_lighting(fggl::entity::EntityManager& ecs) {
		auto light = ecs.create();
		ecs.add<fggl::math::Transform>(light);

		auto& lightComp = ecs.add<fggl::gfx::DirectionalLight>(light);
		lightComp.position = fggl::math::vec3( 10.0F, 5.0F, 0.0F );
		lightComp.diffuse = fggl::gfx::colours::CORNSILK;
		lightComp.ambient = fggl::gfx::colours::MIDNIGHT_BLUE;
		lightComp.specular = fggl::gfx::colours::MIDNIGHT_BLUE;
	}

	Viewer::Viewer(fggl::App &app) : fggl::scenes::Game(app), m_model(fggl::entity::INVALID) {

	}

	void Viewer::activate() {
		Game::activate();

		auto* assetFinder =  m_owner.service<fggl::assets::CheckinAdapted>();
		assetFinder->discover("viewer");

		// setup the assets we can select between
		// TODO some form of introspection to automatically find declared/discovered assets of a given type
		m_assets.clear();
		m_assets.push_back( fggl::assets::make_asset_id_rt("viewer", "backpack/backpack.obj") );
		m_assets.push_back( fggl::assets::make_asset_id_rt("viewer", "lowpoly_scifi/wallDoor_double.FBX") );
		m_assets.push_back( fggl::assets::make_asset_id_rt("viewer", "lowpoly_scifi/wallDoor_double_end.FBX") );
		m_assets.push_back( fggl::assets::make_asset_id_rt("viewer", "newell_teaset/teapot.obj") );
		m_assets.push_back( fggl::assets::make_asset_id_rt("viewer", "humansanimatedpack/Paladin/Paladin.fbx") );

		// create camera
		setup_camera(world());
		setup_lighting(world());

		// setup model
		m_model = fggl::entity::INVALID;
		cycleAsset(0);
	}

	void Viewer::deactivate() {
		Game::deactivate();
	}

	void Viewer::update(float dt) {
		Game::update(dt);
		process_camera(world(), input());

		if ( input().keyboard.pressed(glfwGetKeyScancode(GLFW_KEY_F2)) ) {
			m_debug = !m_debug;
		}

		if ( input().keyboard.pressed(glfwGetKeyScancode(GLFW_KEY_F3)) ) {
			// trigger the asset cycle
			m_lastAsset = (m_lastAsset + 1) % m_assets.size();
			cycleAsset(m_lastAsset);
		}
	}

	void Viewer::cycleAsset(uint64_t /*idx*/) {
		auto *loader = owner().service<fggl::assets::Loader>();
		auto *manager = owner().service<fggl::assets::AssetManager>();

		auto nextAsset = m_assets[ m_lastAsset ];
		loader->loadChain(nextAsset, manager);

		if ( m_model != fggl::entity::INVALID) {
			world().destroy(m_model);
		}
		m_model = build_model(world(), manager, nextAsset);
	}

	void Viewer::render(fggl::gfx::Graphics &gfx) {
		Game::render(gfx);

		gfx.drawScene(world(), m_debug);
	}

}
